package sra.jun.geometry.transformations;

import java.io.*;
import sra.jun.collections.sequences.JunDoubleMatrix;
import sra.jun.geometry.basic.*;

/**
 * Jun2dTransformation class
 * 
 *  @author    nisinaka
 *  @created   1998/12/01 (by nisinaka)
 *  @updated   2000/01/06 (by nisinaka)
 *  @version   445 (with StPL6.5) based on JunXXX for Smalltalk
 *  @copyright 1999-2004 SRA (Software Research Associates, Inc.)
 *  @copyright 1999-2004 Information-technology Promotion Agency, Japan (IPA)
 *  @copyright 2001-2004 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: Jun2dTransformation.java,v 6.7 2004/04/15 06:27:53 nisinaka Exp $
 */
public class Jun2dTransformation extends JunTransformation {
	public double a;
	public double b;
	public double p;
	public double c;
	public double d;
	public double q;
	public double m;
	public double n;
	public double s;

	/**
	 * Create a new instance of Jun2dTransformation and initialize it with an
	 * array.
	 * 
	 * @return sra.jun.geometry.transformations.Jun2Transformation
	 * @param anArray double[]
	 * @category Instance creation
	 */
	public static Jun2dTransformation FromArray_(double[] anArray) {
		Jun2dTransformation aTransformation = new Jun2dTransformation();
		aTransformation.a = anArray[0];
		aTransformation.b = anArray[1];
		aTransformation.p = anArray[2];
		aTransformation.c = anArray[3];
		aTransformation.d = anArray[4];
		aTransformation.q = anArray[5];
		aTransformation.m = anArray[6];
		aTransformation.n = anArray[7];
		aTransformation.s = anArray[8];

		return aTransformation;
	}

	/**
	 * Typical transformation - Rotate around the z axis.
	 * 
	 * @return sra.jun.geometry.transformations.Jun2dTransformation
	 * @param anAngle sra.jun.geometry.basic.JunAngle
	 * @category Typical transformation
	 */
	public static Jun2dTransformation Rotate_(JunAngle anAngle) {
		return RotateZ_(anAngle);
	}

	/**
	 * Typical transformation - Rotate around the z axis.
	 * 
	 * @return sra.jun.geometry.transformations.Jun2dTransformation
	 * @param anAngle sra.jun.geometry.basic.JunAngle
	 * @category Typical transformation
	 */
	public static Jun2dTransformation RotateZ_(JunAngle anAngle) {
		double cos = anAngle.cos();
		double sin = anAngle.sin();
		Jun2dTransformation aTransformation = Unity();
		aTransformation.a = cos;
		aTransformation.b = sin;
		aTransformation.c = -sin;
		aTransformation.d = cos;

		return aTransformation;
	}

	/**
	 * Typical transformation - Rotate around the z axis.
	 * 
	 * @return sra.jun.geometry.transformations.Jun2dTransformation
	 * @param anAngle double
	 * @category Typical transformation
	 */
	public static Jun2dTransformation RotateZ_(double anAngle) {
		return Jun2dTransformation.RotateZ_(JunAngle.FromRad_(anAngle));
	}

	/**
	 * Create a new instance of JunTransformation with the specified scale.
	 * 
	 * @return sra.jun.geometry.transformations.Jun2dTransformation
	 * @param aNumber double
	 * @category Typical transformation
	 */
	public static Jun2dTransformation Scale_(double aNumber) {
		Jun2dTransformation aTransformation = Unity();
		aTransformation.a = aNumber;
		aTransformation.d = aNumber;

		return aTransformation;
	}

	/**
	 * Create a new instance of Jun2dTransformation with the specified scale.
	 * 
	 * @return sra.jun.geometry.transformations.Jun2dTransformation
	 * @param aPoint sra.jun.geometry.basic.Jun2dPoint
	 * @category Typical transformation
	 */
	public static Jun2dTransformation Scale_(Jun2dPoint aPoint) {
		Jun2dTransformation aTransformation = Unity();
		aTransformation.a = aPoint.x();
		aTransformation.d = aPoint.y();

		return aTransformation;
	}

	/**
	 * Create a new instance of Jun2dTransformation with the specified
	 * translation.
	 * 
	 * @return sra.jun.geometry.transformations.Jun2dTransformation
	 * @param aNumber double
	 * @category Typical transformation
	 */
	public static Jun2dTransformation Translate_(double aNumber) {
		Jun2dTransformation aTransformation = Unity();
		aTransformation.m = aNumber;
		aTransformation.n = aNumber;

		return aTransformation;
	}

	/**
	 * Create a new instance of Jun2dTransformation with the specified
	 * translation.
	 * 
	 * @return sra.jun.geometry.transformations.Jun2dTransformation
	 * @param aPoint sra.jun.geometry.basic.Jun2dPoint
	 * @category Typical transformation
	 */
	public static Jun2dTransformation Translate_(Jun2dPoint aPoint) {
		Jun2dTransformation aTransformation = Unity();
		aTransformation.m = aPoint.x();
		aTransformation.n = aPoint.y();

		return aTransformation;
	}

	/**
	 * Answer a new Jun2dTransformation with unit values.
	 * 
	 * @return sra.jun.geometry.transformations.Jun2dTransformation
	 * @category Constants access
	 */
	public static Jun2dTransformation Unity() {
		Jun2dTransformation aTransformation = new Jun2dTransformation();
		aTransformation.a = 1;
		aTransformation.b = 0;
		aTransformation.p = 0;
		aTransformation.c = 0;
		aTransformation.d = 1;
		aTransformation.q = 0;
		aTransformation.m = 0;
		aTransformation.n = 0;
		aTransformation.s = 1;
		return aTransformation;
	}

	/**
	 * Answer a new Jun2dTransformation with zero values.
	 * 
	 * @return sra.jun.geometry.transformations.Jun2dTransformation
	 * @category Constants access
	 */
	public static Jun2dTransformation Zero() {
		Jun2dTransformation aTransformation = new Jun2dTransformation();
		aTransformation.a = 0;
		aTransformation.b = 0;
		aTransformation.p = 0;
		aTransformation.c = 0;
		aTransformation.d = 0;
		aTransformation.q = 0;
		aTransformation.m = 0;
		aTransformation.n = 0;
		aTransformation.s = 0;
		return aTransformation;
	}

	/**
	 * Apply this transformation to the Jun2dPoint and return the result.
	 * 
	 * @return sra.jun.geometry.basic.Jun2dPoint
	 * @param aPoint sra.jun.geometry.basic.Jun2dPoint
	 * @category applying transformation
	 */
	public Jun2dPoint applyToPoint_(Jun2dPoint aPoint) {
		double x = aPoint.x();
		double y = aPoint.y();
		double nx = (x * a) + (y * c) + m;
		double ny = (x * b) + (y * d) + n;
		double scale = (x * p) + (y * q) + s;

		if (scale == 0) {
			scale = ACCURACY;
		}

		return new Jun2dPoint(nx / scale, ny / scale);
	}

	/**
	 * Convert this transformation as an array.
	 * 
	 * @return double[]
	 * @category converting
	 */
	public double[] asArray() {
		double[] array = new double[9];
		array[0] = a;
		array[1] = b;
		array[2] = p;
		array[3] = c;
		array[4] = d;
		array[5] = q;
		array[6] = m;
		array[7] = n;
		array[8] = s;

		return array;
	}

	/**
	 * Convert this transformation as a matrix.
	 * 
	 * @return sra.jun.collections.sequences.JunDoubleMatrix
	 * @category converting
	 */
	public JunDoubleMatrix asMatrix() {
		return new JunDoubleMatrix(3, this.asArray());
	}

	/**
	 * Answer true if aTransformation is equal to this point, otherwise false.
	 * 
	 * @return boolean
	 * @param anObject java.lang.Object
	 * @category comparing
	 */
	public boolean equals(Object anObject) {
		if (anObject instanceof Jun2dTransformation == false) {
			return false;
		}

		Jun2dTransformation aTransformation = (Jun2dTransformation) anObject;

		return (this.a == aTransformation.a)
			&& (this.b == aTransformation.b)
			&& (this.p == aTransformation.p)
			&& (this.c == aTransformation.c)
			&& (this.d == aTransformation.d)
			&& (this.q == aTransformation.q)
			&& (this.m == aTransformation.m)
			&& (this.n == aTransformation.n)
			&& (this.s == aTransformation.s);
	}

	/**
	 * Answer the inverse of this transformation.
	 * 
	 * @return sra.jun.geometry.transformations.Jun2dTransformation
	 * @category transforming functions
	 */
	public Jun2dTransformation inverse() {
		return Jun2dTransformation.FromArray_(this.asMatrix().inverse().asArrayOfDouble());
	}

	/**
	 * Print my string representation on aWriter.
	 * 
	 * @param aWriter java.io.Writer
	 * @throws IOException DOCUMENT ME!
	 * @category printing
	 */
	public void printOn_(Writer aWriter) throws IOException {
		aWriter.write("2dTransformation (");
		aWriter.write(a + " " + b + " " + p);
		aWriter.write(" ,  ");
		aWriter.write(c + " " + d + " " + q);
		aWriter.write(" ,  ");
		aWriter.write(m + " " + n + " " + s);
		aWriter.write(")");
	}

	/**
	 * Answer a new transformation producted by another instance of
	 * Jun2dTransformation.
	 * 
	 * @return sra.jun.geometry.transformations.Jun2dTransformation
	 * @param aTransformation sra.jun.geometry.transformations.Jun2dTransformation
	 * @category transforming functions
	 */
	public Jun2dTransformation product_(Jun2dTransformation aTransformation) {
		Jun2dTransformation newTransformation = new Jun2dTransformation();
		newTransformation.a = (a * aTransformation.a) + (b * aTransformation.c) + (p * aTransformation.m);
		newTransformation.b = (a * aTransformation.b) + (b * aTransformation.d) + (p * aTransformation.n);
		newTransformation.p = (a * aTransformation.p) + (b * aTransformation.q) + (p * aTransformation.s);
		newTransformation.c = (c * aTransformation.a) + (d * aTransformation.c) + (q * aTransformation.m);
		newTransformation.d = (c * aTransformation.b) + (d * aTransformation.d) + (q * aTransformation.n);
		newTransformation.q = (c * aTransformation.p) + (d * aTransformation.q) + (q * aTransformation.s);
		newTransformation.m = (m * aTransformation.a) + (n * aTransformation.c) + (s * aTransformation.m);
		newTransformation.n = (m * aTransformation.b) + (n * aTransformation.d) + (s * aTransformation.n);
		newTransformation.s = (m * aTransformation.p) + (n * aTransformation.q) + (s * aTransformation.s);
		return newTransformation;
	}
}
