package sra.jun.opengl.flux;

import java.awt.Dimension;
import java.awt.Image;
import java.awt.event.WindowEvent;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.text.DateFormat;
import java.util.Date;
import sra.smalltalk.SmalltalkException;
import sra.smalltalk.StBlockClosure;
import sra.smalltalk.StImage;
import sra.smalltalk.StMenu;
import sra.smalltalk.StMenuItem;
import sra.smalltalk.StMenuModel;
import sra.smalltalk.StSymbol;
import sra.smalltalk.StView;
import sra.jun.geometry.basic.Jun3dBoundingBox;
import sra.jun.geometry.basic.Jun3dPoint;
import sra.jun.goodies.animation.JunCartoonMovie;
import sra.jun.goodies.cursors.JunCursors;
import sra.jun.goodies.files.JunFileDialog;
import sra.jun.goodies.files.JunFileModel;
import sra.jun.goodies.images.streams.JunImageStream;
import sra.jun.goodies.images.streams.JunJpegImageStream;
import sra.jun.goodies.movie.JunImagesToMovie;
import sra.jun.goodies.movie.JunMoviePlayer;
import sra.jun.goodies.track.JunTrackerModel;
import sra.jun.opengl.display.JunOpenGLDisplayModel;
import sra.jun.opengl.objects.JunOpenGL3dCompoundObject;
import sra.jun.opengl.projection.JunOpenGLProjection;
import sra.jun.system.framework.JunDialog;
import sra.jun.system.support.JunSystem;

/**
 * JunOpenGLFluxModel class
 * 
 *  @author    MATSUDA Ryouichi
 *  @created   1998/11/25 (by MATSUDA Ryouichi)
 *  @updated   1999/06/25 (by nisinaka)
 *  @updated   2003/03/24 (by nisinaka)
 *  @version   445 (with StPL6.5) based on Jun490 for Smalltalk
 *  @copyright 1999-2004 SRA (Software Research Associates, Inc.)
 *  @copyright 1999-2004 Information-technology Promotion Agency, Japan (IPA)
 *  @copyright 2001-2004 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: JunOpenGLFluxModel.java,v 6.10 2004/04/15 06:28:08 nisinaka Exp $
 */
public class JunOpenGLFluxModel extends JunOpenGLDisplayModel {

	protected JunOpenGLFluxObject fluxObject;
	protected JunTrackerModel trackerModel;

	/**
	 * Create a new instance of JunOpenGLFluxModel and initialize it.
	 * 
	 * @category Instance creation
	 */
	public JunOpenGLFluxModel() {
		super();
	}

	/**
	 * Create a new instance of JunOpenGLFluxModel and initialize it with a flux object.
	 * 
	 * @param aFluxObject sra.jun.opengl.flux.JunOpenGLFluxObject
	 * @category Instance creation
	 */
	public JunOpenGLFluxModel(JunOpenGLFluxObject aFluxObject) {
		this();
		this.fluxObject_(aFluxObject);
	}

	/**
	 * Initialize the JunOpenGLFluxModel.
	 * 
	 * @see sra.smalltalk.StApplicationModel#initialize()
	 * @category initialize-release
	 */
	protected void initialize() {
		super.initialize();
		fluxObject = null;
		trackerModel = null;
	}

	/**
	 * Release the JunOpenGLFluxModel.
	 * 
	 * @category initialize-release
	 */
	public void release() {
		this.trackerModel().end();
	}

	/**
	 * Answer my current flux object.
	 * 
	 * @return sra.jun.opengl.flux.JunOpenGLFluxObject
	 * @category accessing
	 */
	public JunOpenGLFluxObject fluxObject() {
		return fluxObject;
	}

	/**
	 * Set my new flux object.
	 * 
	 * @param aFluxObject sra.jun.opengl.flux.JunOpenGLFluxObject
	 * @category accessing
	 */
	public void fluxObject_(JunOpenGLFluxObject aFluxObject) {
		fluxObject = aFluxObject;

		if (fluxObject.size() - 1 <= 0) {
			this.trackerModel().step_(0.0d);
		} else {
			this.trackerModel().step_(1.0d / (fluxObject.size() - 1));
		}
		this.trackerModel().first();
	}

	/**
	 * Answer my tracker model.
	 * 
	 * @return sra.jun.goodies.track.JunTrackerModel
	 * @category accessing
	 */
	public JunTrackerModel trackerModel() {
		if (trackerModel == null) {
			trackerModel = new JunTrackerModel();
			trackerModel.compute_(new StBlockClosure() {
				public Object value_(Object anObject) {
					if (JunOpenGLFluxModel.this.fluxObject() != null) {
						double n = ((Number) anObject).doubleValue();
						JunOpenGLFluxModel.this.displayObject_(fluxObject.at_(n));
						JunOpenGLFluxModel.this.changed_($("object"));
						Thread.yield();
					}
					return null;
				}
			});
		}

		return trackerModel;
	}

	/**
	 * Answer my current flux object as JunOpenGL3dCompoundObject.
	 * 
	 * @return sra.jun.opengl.objects.JunOpenGL3dCompoundObject
	 * @category accessing
	 */
	public JunOpenGL3dCompoundObject compoundObject() {
		return this.fluxObject().compoundObject();
	}

	/**
	 * Start the tracker model.
	 * 
	 * @category accessing
	 */
	public void start() {
		this.trackerModel().start();
	}

	/**
	 * End the tracking process.
	 * 
	 * @category accessing
	 */
	public void end() {
		this.trackerModel().end();
	}

	/**
	 * Answer the bounding box of the receiver.
	 * 
	 * @return sra.jun.geometry.basic.Jun3dBoundingBox
	 * @category bounds accessing
	 */
	public Jun3dBoundingBox boundingBox() {
		if (this.fluxObject() == null) {
			return Jun3dBoundingBox.Origin_corner_(new Jun3dPoint(0, 0, 0), new Jun3dPoint(0, 0, 0));
		}

		return this.fluxObject().boundingBox();
	}

	/**
	 * Enumerate the points and evaluate the Block.
	 * 
	 * @param aBlock sra.smalltalk.StBlockClosure
	 * @category enumerating
	 */
	public void pointsDo_(final StBlockClosure aBlock) {
		if (this.fluxObject() == null) {
			this.pointsDo_(aBlock);
		} else {
			this.fluxObject().pointsDo_(aBlock);
		}
	}

	/**
	 * Answer a window title.
	 * 
	 * @return java.lang.String
	 * @cagtegory interface opening
	 */
	protected String windowTitle() {
		return JunSystem.$String("Flux");
	}

	/**
	 * Update the indication of the file menu.
	 * 
	 * @category menu accessing
	 */
	public void updateFileMenuIndication() {
		super.updateFileMenuIndication();

		StMenu fileMenu = this._fileMenu();
		if (fileMenu == null) {
			return;
		}

		StMenuItem menuItem;
		boolean displayObjectIsNotEmpty = !this.isEmpty();

		// Save...
		menuItem = fileMenu.getItemWithValue("saveLFT");
		if (menuItem != null) {
			menuItem.setEnabled(displayObjectIsNotEmpty);
		}

		// Save as images...
		menuItem = fileMenu.getItemWithValue("saveAsImages");
		if (menuItem != null) {
			menuItem.setEnabled(displayObjectIsNotEmpty);
		}

		// Save as movie...
		menuItem = fileMenu.getItemWithValue("saveAsMovie");
		if (menuItem != null) {
			menuItem.setEnabled(displayObjectIsNotEmpty);
		}
	}

	/**
	 * Update the indication of the misc menu.
	 * 
	 * @category menu accessing
	 */
	public void updateMiscMenuIndication() {
		super.updateMiscMenuIndication();

		StMenu miscMenu = this._miscMenu();
		if (miscMenu == null) {
			return;
		}

		StMenuItem menuItem;
		boolean displayObjectIsNotEmpty = !this.isEmpty();

		// Convert to animcation...
		menuItem = miscMenu.getItemWithValue("convertToAnimation");
		if (menuItem != null) {
			menuItem.setEnabled(displayObjectIsNotEmpty);
		}
	}

	/**
	 * Called by menu of 'Open...'
	 * 
	 * @category menu accessing
	 */
	public void openLFT() {
		this.openLFT10();
	}

	/**
	 * Open an LFT1.0 file and get a JunOpenGLFluxObject.
	 * 
	 * @throws sra.smalltalk.SmalltalkException
	 * @category menu messages
	 */
	public void openLFT10() {
		JunFileModel.FileType[] fileTypes = new JunFileModel.FileType[] { new JunFileModel.FileType(JunSystem.$String("<1p> files", null, "LFT"), new String[] { "*.lft", "*.LFT" })
		};
		File aFile = JunFileDialog.Request(JunSystem.$String("Select an <1p> file.", null, "LFT"), fileTypes, fileTypes[0]);
		if (aFile == null) {
			return;
		}

		JunOpenGLFluxObject anObject;
		try {
			anObject = this.readFromLFT10_(aFile);
		} catch (IOException e) {
			throw new SmalltalkException(e);
		}
		if (anObject == null) {
			return;
		}

		this.fluxObject_(anObject);
		this.resetView();

		if (this.showModel() != null) {
			this.showModel().resetView();
		}
	}

	/**
	 * Called by menu of 'Save...'
	 * 
	 * @category menu messages
	 */
	public void saveLFT() {
		this.saveLFT10();
	}

	/**
	 * Save the flux object to a file.
	 * 
	 * @throws sra.smalltalk.SmalltalkException
	 * @category menu messages
	 */
	public void saveLFT10() {
		JunFileModel.FileType[] fileTypes = new JunFileModel.FileType[] { new JunFileModel.FileType(JunSystem.$String("<1p> files", null, "LFT"), new String[] { "*.lft", "*.LFT" })
		};
		File file = JunFileDialog.RequestNewFilename(JunSystem.$String("Input an <1p> file.", null, "LFT"), new File(this.fluxObject().name() + ".lft"), fileTypes, fileTypes[0]);
		if (file == null) {
			return;
		}

		try {
			this.writeToLFT10_object_(file, this.fluxObject());
		} catch (IOException e) {
			throw new SmalltalkException(e);
		}
	}

	/**
	 * Save the receiver as a series of images.
	 * 
	 * @category menu messages
	 */
	public void saveAsImages() {
		if (this.fluxObject() == null) {
			return;
		}

		this.end();

		File directory = JunFileDialog.RequestNewDirectory(JunSystem.$String("Input a directory name."), new File(this.displayObject().name()));
		if (directory == null) {
			return;
		}

		this.saveAsImagesTo_(directory);
	}

	/**
	 * Save the receiver as a series of images.
	 * 
	 * @param directory java.io.File
	 * @category menu messages
	 */
	protected void saveAsImagesTo_(File directory) {
		if (directory.exists() == false) {
			directory.mkdir();
		}
		Thread.yield();

		StSymbol loopCondition = this.trackerModel().loopCondition();
		double currentValue = this.trackerModel().doubleValue();
		this.trackerModel().loopCondition_($("oneWay"));
		JunCursors cursor = new JunCursors(JunCursors.ExecuteCursor());
		try {
			cursor._show();

			this.trackerModel().playButtonVisual_(true);
			this.trackerModel().first();
			for (int index = 1; this.trackerModel().playButton().value(); index++) {
				String aString = "0000" + index;
				aString = aString.substring(aString.length() - 4);
				File aFile = new File(directory, aString + ".jpg");
				Image anImage = this.asImage();

				JunCursors writeCursor = new JunCursors(JunCursors.WriteCursor());
				JunImageStream aStream = null;
				try {
					writeCursor._show();
					aStream = JunJpegImageStream.On_(new FileOutputStream(aFile));
					aStream.nextPutImage_(new StImage(anImage));
				} catch (IOException e) {
					throw new SmalltalkException(e);
				} finally {
					if (aStream != null) {
						try {
							aStream.flush();
							aStream.close();
						} catch (IOException e) {
						} finally {
							aStream = null;
						}
					}
					writeCursor._restore();
				}

				this.trackerModel().next();
				Thread.yield();
			}

			this.trackerModel().playButtonVisual_(false);
		} finally {
			this.trackerModel().loopCondition_(loopCondition);
			this.trackerModel().value_(currentValue);

			cursor._restore();
		}
	}

	/**
	 * Save the receiver as a movie.
	 * 
	 * @category menu messages
	 */
	public void saveAsMovie() {
		if (this.fluxObject() == null) {
			return;
		}

		this.end();

		JunFileModel.FileType[] fileTypes = new JunFileModel.FileType[] { new JunFileModel.FileType(JunSystem.$String("<1p> files", null, JunSystem.$String("Movie")), new String[] { "*.mov", "*.MOV" })
		};
		File file = JunFileDialog.RequestNewFilename(JunSystem.$String("Input a <1p> file.", null, JunSystem.$String("Movie")), new File(this.displayObject().name() + ".mov"), fileTypes, fileTypes[0]);
		if (file == null) {
			return;
		}

		if (this.saveAsMovieTo_(file)) {
			JunMoviePlayer player = new JunMoviePlayer(file);
			if (player != null) {
				player.open();
			}
		}
	}

	/**
	 * Save the receiver as a movie.
	 * 
	 * @param file java.io.File
	 * @category  menu messages
	 */
	protected boolean saveAsMovieTo_(File file) {
		if (file == null) {
			return false;
		}

		Thread.yield();

		StSymbol loopCondition = this.trackerModel().loopCondition();
		double currentValue = this.trackerModel().doubleValue();
		this.trackerModel().loopCondition_($("oneWay"));
		JunCursors cursor = new JunCursors(JunCursors.ExecuteCursor());
		try {
			cursor._show();

			this.trackerModel().playButtonVisual_(true);
			this.trackerModel().first();
			Dimension extent = this.defaultImageExtent();
			(new JunImagesToMovie(file.getAbsolutePath(), extent)).do_(new StBlockClosure() {
				public Object value_(Object imagesToMovie) {
					while (JunOpenGLFluxModel.this.trackerModel().playButton().value()) {
						Image anImage = JunOpenGLFluxModel.this.asImage();
						((JunImagesToMovie) imagesToMovie).add_(new StImage(anImage));
						JunOpenGLFluxModel.this.trackerModel().next();
						Thread.yield();
					}
					return null;
				}
			});
			this.trackerModel().playButtonVisual_(false);
		} finally {
			this.trackerModel().loopCondition_(loopCondition);
			this.trackerModel().value_(currentValue);

			cursor._restore();
		}

		return true;
	}

	/**
	 * Called by menu of 'spawn'
	 * 
	 * @category menu messages
	 */
	public void spawnObject() {
		JunOpenGLDisplayModel displayModel = new JunOpenGLDisplayModel(this.spawningObject());
		displayModel.displayProjection_((JunOpenGLProjection) this.displayProjection().copy());
		displayModel.open();
		displayModel.changed_($("object"));
	}

	/**
	 * Convert the receiver to the animation.
	 * 
	 * @category menu messages
	 */
	public void convertToAnimation() {
		String aString = JunDialog.Request_(JunSystem.$String("Input the interval milliseconds."), String.valueOf(100));
		if (aString == null | aString.length() == 0) {
			return;
		}

		try {
			int intervalMilliseconds = Integer.parseInt(aString);
			this.convertToAnimation_(intervalMilliseconds);
		} catch (NumberFormatException e) {
			JunDialog.Warn_(aString + JunSystem.$String(" is invalid value."));
			return;
		}
	}

	/**
	 * Convert the receiver to the animation.
	 * 
	 * @param intervalMilliseconds int
	 * @category menu messages
	 */
	protected void convertToAnimation_(int intervalMilliseconds) {
		JunCartoonMovie cartoonMovie = new JunCartoonMovie();
		int millisecondValue = Math.max(intervalMilliseconds, cartoonMovie.tickTime());
		File directory = new File(this.defaultBaseName());
		this.saveAsImagesTo_(directory);

		File[] files = directory.listFiles();
		try {
			for (int i = 0; i < files.length; i++) {
				JunImageStream aStream = null;
				try {
					aStream = JunJpegImageStream.On_(new FileInputStream(files[i]));
					StImage anImage = aStream.nextImage();
					cartoonMovie.addImage_keepTime_(anImage, millisecondValue);
				} finally {
					if (aStream != null) {
						aStream.close();
						aStream = null;
					}
				}
			}
		} catch (IOException e) {
			throw new SmalltalkException(e);
		} finally {
			for (int i = 0; i < files.length; i++) {
				files[i].delete();
			}
			directory.delete();
		}

		cartoonMovie.open();
	}

	/**
	 * Invoked when a window is in the process of being closed.
	 * 
	 * @param e java.awt.event.WindowEvent
	 * @see sra.smalltalk.StApplicationModel#noticeOfWindowClose(java.awt.event.WindowEvent)
	 * @category events
	 */
	public void noticeOfWindowClose(WindowEvent e) {
		super.noticeOfWindowClose(e);
		this.release();
	}

	/**
	 * Read a LFT1.0 file and create a flux object.
	 * 
	 * @param aFile java.io.File
	 * @return sra.jun.opengl.flux.JunOpenGLFluxObject
	 * @exception IOException
	 * @category reading
	 */
	public JunOpenGLFluxObject readFromLFT10_(File aFile) throws IOException {
		FileReader aFileReader;
		try {
			aFileReader = new FileReader(aFile);
		} catch (FileNotFoundException e) {
			return null;
		}

		BufferedReader aReader = new BufferedReader(aFileReader);
		JunOpenGLFluxObject anObject = null;
		try {
			anObject = this.loadFromLFT10_(aReader);
		} finally {
			aReader.close();
		}

		return anObject;
	}

	/**
	 * Load a JunOpenGLFluxObject from the reader.
	 * 
	 * @param aReader java.io.BufferedReader
	 * @return sra.jun.opengl.flux.JunOpenGLFluxObject
	 * @exception IOException
	 * @category reading
	 */
	public JunOpenGLFluxObject loadFromLFT10_(BufferedReader aReader) throws IOException {
		StringWriter aWriter = new StringWriter();
		JunOpenGLFluxObject anObject = null;
		try {
			int ch;
			while ((ch = aReader.read()) > 0) {
				aWriter.write(ch);
			}
			aWriter.flush();
			anObject = JunOpenGLFluxObject.LoadFrom_(aWriter.toString());
		} finally {
			aWriter.close();
		}

		return anObject;
	}

	/**
	 * Write the JunOpenGLFluxObject to the file.
	 * 
	 * @param aFile java.io.File
	 * @param anObject sra.jun.opengl.flux.JunOpenGLFluxObject
	 * @exception java.io.IOException
	 * @category writing
	 */
	public void writeToLFT10_object_(File aFile, JunOpenGLFluxObject anObject) throws IOException {
		if (aFile == null) {
			return;
		}

		BufferedWriter aWriter = new BufferedWriter(new FileWriter(aFile));
		try {
			this.saveToLFT10_object_(aWriter, anObject);
		} finally {
			aWriter.flush();
			aWriter.close();
		}
	}

	/**
	 * Write the JunOpenGLFluxObject to the writer.
	 * 
	 * @param aWriter java.io.BufferedWriter
	 * @param anObject sra.jun.opengl.flux.JunOpenGLFluxObject
	 * @exception java.io.IOException
	 * @category writing
	 */
	public void saveToLFT10_object_(BufferedWriter aWriter, JunOpenGLFluxObject anObject) throws IOException {
		aWriter.write(this.defaultStampForLFT10());
		anObject.saveOn_(aWriter);
	}

	/**
	 * Answer a default view.
	 * 
	 * @return sra.smalltalk.StView
	 * @category defaults
	 */
	public StView defaultView() {
		if (GetDefaultViewMode() == VIEW_AWT) {
			return new JunOpenGLFluxViewAwt(this);
		} else {
			return new JunOpenGLFluxViewSwing(this);
		}
	}

	/**
	 * Answer the stamp string for the LFT 1.0 file format.
	 * 
	 * @return java.lang.String
	 * @category defaults
	 */
	public String defaultStampForLFT10() {
		StringWriter sw = new StringWriter();
		PrintWriter pw = new PrintWriter(sw);
		pw.println("% LFT V1.0 List Flux Transmission (Lisp S Expression)");
		pw.println("% This file was created by " + JunSystem.System() + JunSystem.Version());
		pw.println("% " + DateFormat.getInstance().format(new Date()));
		pw.println();
		pw.flush();
		return sw.toString();
	}

	/**
	 * Create the menubar for JunOpenGLDisplay.
	 * 
	 * @return sra.smalltalk.StMenuModel
	 * @category defaults
	 */
	protected StMenuModel defaultMenu() {
		StMenuModel rootMenu = new StMenuModel();
		rootMenu.add(this._fileMenu());
		rootMenu.add(this._editMenu());
		rootMenu.add(this._viewMenu());
		rootMenu.add(this._lightMenu());
		rootMenu.add(this._miscMenu());
		return rootMenu;
	}

	/**
	 * Answer the default file menu for JunOpenGLFluxModel.
	 * 
	 * @return sra.smalltalk.StMenu
	 * @category defaults
	 */
	protected StMenu _defaultFileMenu() {
		StMenu defaultFileMenu = new StMenu(JunSystem.$String("File"));
		defaultFileMenu.add(createMenuItem(JunSystem.$String("New"), "newModel"));
		defaultFileMenu.add(createMenuItem(JunSystem.$String("Open") + "...", "openLFT"));
		defaultFileMenu.addSeparator();
		defaultFileMenu.add(createMenuItem(JunSystem.$String("Save") + "...", "saveLFT"));
		defaultFileMenu.add(createMenuItem(JunSystem.$String("Save as image..."), "saveAsImage"));
		defaultFileMenu.add(createMenuItem(JunSystem.$String("Save as images..."), "saveAsImages"));
		defaultFileMenu.add(createMenuItem(JunSystem.$String("Save as movie..."), "saveAsMovie"));
		defaultFileMenu.addSeparator();
		defaultFileMenu.add(createMenuItem(JunSystem.$String("Quit"), "quitDoing"));
		return defaultFileMenu;
	}

	/**
	 * Answer the default misc menu for JunOpenGLFluxModel.
	 * 
	 * @return sra.smalltalk.StMenu
	 * @category defaults
	 */
	protected StMenu _defaultMiscMenu() {
		StMenu defaultMiscMenu = new StMenu(JunSystem.$String("Misc"));
		defaultMiscMenu.add(createMenuItem(JunSystem.$String("Spawn"), "spawnObject"));
		defaultMiscMenu.add(createMenuItem(JunSystem.$String("Viewport"), "spawnViewport"));
		defaultMiscMenu.add(createMenuItem(JunSystem.$String("Bounds"), "showBounds"));
		defaultMiscMenu.add(createMenuItem(JunSystem.$String("Convert to animation..."), "convertToAnimation"));
		return defaultMiscMenu;
	}

}
