package sra.jun.terrain.display;

import java.awt.*;
import sra.smalltalk.*;
import sra.jun.geometry.basic.*;
import sra.jun.geometry.curves.Jun3dLine;
import sra.jun.geometry.transformations.Jun3dTransformation;
import sra.jun.goodies.button.JunButtonModel;
import sra.jun.goodies.cursors.JunCursors;
import sra.jun.opengl.cosmos.JunOpenGLMicrocosmModel;
import sra.jun.opengl.display.*;
import sra.jun.opengl.objects.JunOpenGL3dObject;
import sra.jun.opengl.projection.*;
import sra.jun.system.framework.JunDialog;
import sra.jun.system.support.JunSystem;
import sra.jun.terrain.support.JunTerrainField;

/**
 * JunTerrainDisplayModel class
 * 
 *  @author    nisinaka
 *  @created   2002/01/22 (by nisinaka)
 *  @updated   2003/05/15 (by nisinaka)
 *  @version   445 (with StPL6.5) based on Jun470 for Smalltalk
 *  @copyright 1999-2004 SRA (Software Research Associates, Inc.)
 *  @copyright 1999-2004 Information-technology Promotion Agency, Japan (IPA)
 *  @copyright 2001-2004 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: JunTerrainDisplayModel.java,v 6.7 2004/04/15 06:28:13 nisinaka Exp $
 */
public class JunTerrainDisplayModel extends JunOpenGLMicrocosmModel {

	protected JunTerrainField terrainField;
	protected double inches;
	protected double defaultZoomHeight;

	/**
	 * Create a new instance of JunTerrainDisplayModel to show the terrain
	 * field.
	 * 
	 * @param aTerrainField sra.jun.terrain.support.JunTerrainField
	 */
	public JunTerrainDisplayModel(JunTerrainField aTerrainField) {
		terrainField = aTerrainField;
	}

	/**
	 * menu message: change inches.
	 */
	public void changeInches() {
		String numberString = JunDialog.Request_(JunSystem.$String("Input your inches."), String.valueOf(this.inches()));
		if (numberString == null) {
			return;
		}

		try {
			double newInches = Double.parseDouble(numberString);
			this.inches_(newInches);
		} catch (NumberFormatException e) {
			JunDialog.Warn_(numberString + JunSystem.$String(" is invalid value."));
		}
	}

	/**
	 * Answer the default light color.
	 * 
	 * @return java.awt.Color
	 */
	public Color defaultLightColor() {
		return Color.getHSBColor(0.19f, 0.5f, 1.0f);
	}

	/**
	 * Answer the default light point.
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 */
	public Jun3dPoint defaultLightPoint() {
		return new Jun3dPoint(0, 0.5, 1);
	}

	/**
	 * Answer the default view.
	 * 
	 * @return sra.smalltalk.StView
	 */
	public StView defaultView() {
		StView aView = super.defaultView();

		// Set the controller.
		aView.controller_(this.defaultController());

		((Component) ((JunOpenGLDisplayView) aView).getOpenGLView()).setBackground(Color.getHSBColor(0.5f, 0.5f, 1.0f));

		return aView;
	}

	/**
	 * Answer the default view factor.
	 * 
	 * @return double
	 */
	public double defaultViewFactor() {
		return 5.0d;
	}

	/**
	 * Answer the current default zoom height.
	 * 
	 * @return double
	 */
	public double defaultZoomHeight() {
		return defaultZoomHeight;
	}

	/**
	 * Set the new default zoom height.
	 * 
	 * @param newDefaultZoomHeight double
	 */
	public void defaultZoomHeight_(double newDefaultZoomHeight) {
		defaultZoomHeight = newDefaultZoomHeight;
	}

	/**
	 * Answer the array of all display lights.
	 * 
	 * @return sra.jun.opengl.display.JunOpenGLDisplayLight[]
	 */
	public JunOpenGLDisplayLight[] displayLights() {
		if (displayLights == null) {
			displayLights = new JunOpenGLDisplayLight[5];
			displayLights[0] = JunOpenGLDisplayLight.ParallelLight_color_position_(true, this.defaultLightColor(), this.defaultLightPoint());
			displayLights[1] = new JunOpenGLDisplayLight();
			displayLights[2] = new JunOpenGLDisplayLight();
			displayLights[3] = new JunOpenGLDisplayLight();
			displayLights[4] = JunOpenGLDisplayLight.AmbientLight_color_(false, this.defaultLightColor());

			for (int i = 0; i < displayLights.length; i++) {
				displayLights[i].compute_(new StBlockClosure() {
					public Object value() {
						updateLightMenuIndication();
						changed_($("light"));

						return null;
					}
				});
			}
		}

		return displayLights;
	}

	/**
	 * Answer the current display object.
	 * 
	 * @return sra.jun.opengl.objects.JunOpenGL3dObject
	 */
	public JunOpenGL3dObject displayObject() {
		if (openGL3dObject == null) {
			openGL3dObject = this.computeDisplayObject();
		}

		return openGL3dObject;
	}

	/**
	 * Answer my display projector.
	 * 
	 * @return sra.jun.opengl.projection.JunOpenGLProjector
	 */
	public JunOpenGLProjector displayProjector() {
		if (openGLProjector == null) {
			openGLProjector = new JunOpenGLProjector();
			openGLProjector.projection_(JunTerrainProjection.Default());
			openGLProjector.eyePoint_(this.defaultEyePoint());
			openGLProjector.sightPoint_(this.defaultSightPoint());
			openGLProjector.upVector_(this.defaultUpVector());
			openGLProjector.viewFactor_(this.defaultViewFactor());
			openGLProjector.far_(this.defaultFar());
			openGLProjector.near_(this.defaultNear());
		}

		return openGLProjector;
	}

	/**
	 * Dolly with the specified factor.
	 * 
	 * @param factor double
	 */
	public void dolly_(double factor) {
		double dollyFactor = factor;
		JunOpenGLProjection projection = this.displayProjection();
		Jun3dPoint dollyVector = projection.sightVector().negated().multipliedBy_(dollyFactor);
		Jun3dPoint eyePoint = projection.eyePoint();
		this.moveTo_(new Jun2dPoint(eyePoint.x() + dollyVector.x(), eyePoint.y() + dollyVector.y()));
	}

	/**
	 * Answer the current dolly button model.
	 * 
	 * @return sra.jun.goodies.button.JunButtonModel
	 */
	public JunButtonModel dollyButton() {
		if (dollyButton == null) {
			JunButtonModel button = new JunButtonModel();
			button.value_(true);
			button.visual_(JunCursors.DollyCursorImage());
			button.action_(new StBlockClosure() {
				public Object value_(Object o) {
					JunButtonModel model = (JunButtonModel) o;
					model.value_(!model.value());

					return model;
				}
			});
			dollyButton = button;
		}

		return dollyButton;
	}

	/**
	 * Fit the 3d object to the current view size.
	 */
	public void fitSilently() {
		if (this.displayObject() == null) {
			return;
		}

		JunOpenGLProjector projector = this.displayProjector();
		projector.sightPoint_(this.defaultSightPoint());
		projector.eyePoint_(this.defaultEyePoint());
		projector.upVector_(this.defaultUpVector());
		projector.viewFactor_(this.defaultViewFactor());
		projector.far_(this.defaultFar());
		projector.near_(this.defaultNear());
		projector.zoomHeight_(this.defaultZoomHeight());
	}

	/**
	 * Grab and rotate the 3d object.
	 * 
	 * @param from2dPoint sra.jun.grometry.basic.Jun2dPoint
	 * @param to2dPoint sra.jun.grometry.basic.Jun2dPoint
	 */
	public void grab_xy_(Jun2dPoint from2dPoint, Jun2dPoint to2dPoint) {
		this.grab_x_(from2dPoint, to2dPoint);
		this.grab_y_(from2dPoint, to2dPoint);
		this.changed_($("projection"));
	}

	/**
	 * Answer the current inches.
	 * 
	 * @return double
	 */
	public double inches() {
		return inches;
	}

	/**
	 * Set the new inches.
	 * 
	 * @param newInches double
	 */
	public void inches_(double newInches) {
		inches = newInches;
		this.moveTo_(this.displayProjector().eyePoint());
	}

	/**
	 * Move your eyes to the specified point with the sight.
	 * 
	 * @param eye2dPoint sra.jun.geometry.basic.Jun2dPoint
	 * @param sight2dPoint sra.jun.geometry.basic.Jun2dPoint
	 */
	public void moveTo_sight_(Jun2dPoint eye2dPoint, Jun2dPoint sight2dPoint) {
		JunOpenGLProjector projector = this.displayProjector();
		Jun3dPoint newEyePoint = new Jun3dPoint(eye2dPoint, this.eyeHeightAtPoint_(eye2dPoint));
		Jun3dPoint newSightPoint = new Jun3dPoint(sight2dPoint, terrainField.zAt2dPoint_(sight2dPoint));
		projector.sightPoint_(newSightPoint);
		projector.eyePoint_(newEyePoint);
		projector.upVector_(new Jun3dPoint(0, 0, 1));
		this.changed_($("projection"));
	}

	/**
	 * menu message: show the display object.
	 */
	public void showDisplayObject() {
		this.displayObject().show();
	}

	/**
	 * menu message: spawn the object.
	 */
	public void spawnObject() {
		JunTerrainDisplayModel displayModel = new JunTerrainDisplayModel(terrainField);
		displayModel.defaultSightPoint_(this.displayProjection().sightPoint());
		displayModel.defaultEyePoint_(this.displayProjection().eyePoint());
		displayModel.defaultZoomHeight_(this.defaultZoomHeight());
		displayModel.defaultUpVector_(new Jun3dPoint(0, 0, 1));
		displayModel.inches_(this.inches());
		displayModel.open();
		displayModel.resetView();
		displayModel.changed_($("object"));
	}

	/**
	 * Answer the default file menu for JunOpenGLDisplayModel.
	 * 
	 * @return sra.smalltalk.StMenu
	 */
	protected StMenu _defaultFileMenu() {
		StMenu defaultFileMenu = new StMenu(JunSystem.$String("File"));

		// Quit
		defaultFileMenu.add(createMenuItem(JunSystem.$String("Quit"), "quitDoing"));

		return defaultFileMenu;
	}

	/**
	 * Answer the default view menu for JunOpenGL3dModel.
	 * 
	 * @return sra.smalltalk.StMenu
	 */
	protected StMenu _defaultViewMenu() {
		StMenu defaultViewMenu = new StMenu(JunSystem.$String("View"));

		// Inches...
		defaultViewMenu.add(createMenuItem(JunSystem.$String("Inch", "Inches") + "...", "changeInches"));

		//
		defaultViewMenu.addSeparator();

		// Zooming
		StMenu zoomingMenu = new StMenu(JunSystem.$String("Zooming"));
		zoomingMenu.add(createMenuItem(JunSystem.$String("Fit zoom"), "fitZoom"));
		zoomingMenu.add(createMenuItem(JunSystem.$String("Close up"), "closeUp"));
		zoomingMenu.addSeparator();
		zoomingMenu.add(createMenuItem(JunSystem.$String("Zoom"), "zoom"));
		zoomingMenu.add(createMenuItem(JunSystem.$String("Pan"), "pan"));
		defaultViewMenu.add(zoomingMenu);

		// Pointing
		StMenu pointingMenu = new StMenu(JunSystem.$String("Pointing"));
		pointingMenu.add(createMenuItem(JunSystem.$String("Fit sight"), "fitSight"));
		pointingMenu.addSeparator();
		pointingMenu.add(createMenuItem(JunSystem.$String("Eye point") + "...", "changeEyePoint"));
		pointingMenu.add(createMenuItem(JunSystem.$String("Sight point") + "...", "changeSightPoint"));
		pointingMenu.add(createMenuItem(JunSystem.$String("Up vector") + "...", "changeUpVector"));
		pointingMenu.addSeparator();
		pointingMenu.add(createMenuItem(JunSystem.$String("Zoom height") + "...", "changeZoomHeight"));
		pointingMenu.add(createMenuItem(JunSystem.$String("View factor") + "...", "changeViewFactor"));
		defaultViewMenu.add(pointingMenu);

		// Projection
		StMenu projectionMenu = new StMenu(JunSystem.$String("Projection"));
		StButtonGroup projectionGroup = new StButtonGroup();
		projectionMenu.add(createRadioButtonMenuItem(JunSystem.$String("Perspective Projection", "Perspective"), "perspective", projectionGroup));
		projectionMenu.add(createRadioButtonMenuItem(JunSystem.$String("Parallel Projection", "Parallel"), "parallelProjection", projectionGroup));
		defaultViewMenu.add(projectionMenu);

		// Presentation
		StMenu presentationMenu = new StMenu(JunSystem.$String("Presentation"));
		StButtonGroup presentationGroup = new StButtonGroup();
		presentationMenu.add(createRadioButtonMenuItem(JunSystem.$String("Solid Presentation", "Solid"), "solidPresentation", presentationGroup));
		presentationMenu.add(createRadioButtonMenuItem(JunSystem.$String("Wireframe Presentation", "Wireframe"), "wireframePresentation", presentationGroup));
		presentationMenu.add(createRadioButtonMenuItem(JunSystem.$String("Hidden-line Presentation", "Hidden-line"), "hiddenlinePresentation", presentationGroup));
		defaultViewMenu.add(presentationMenu);

		// Shading
		StMenu shadingMenu = new StMenu(JunSystem.$String("Shading"));
		StButtonGroup shadingGroup = new StButtonGroup();
		shadingMenu.add(createRadioButtonMenuItem(JunSystem.$String("Flat Shading", "Flat"), "flatShading", shadingGroup));
		shadingMenu.add(createRadioButtonMenuItem(JunSystem.$String("Smooth Shading", "Smooth"), "smoothShading", shadingGroup));
		defaultViewMenu.add(shadingMenu);

		// Smoothing
		StMenu smoothingMenu = new StMenu(JunSystem.$String("Smoothing"));
		smoothingMenu.add(createCheckBoxMenuItem(JunSystem.$String("Line smooth"), "lineSmooth"));
		smoothingMenu.add(createCheckBoxMenuItem(JunSystem.$String("Polygon smooth"), "polygonSmooth"));
		defaultViewMenu.add(smoothingMenu);

		// and others.
		defaultViewMenu.addSeparator();
		defaultViewMenu.add(createMenuItem(JunSystem.$String("Reset"), "resetView"));

		return defaultViewMenu;
	}

	/**
	 * Create the current display object.
	 * 
	 * @return sra.jun.opengl.objects.JunOpenGL3dObject
	 */
	protected JunOpenGL3dObject computeDisplayObject() {
		if (terrainField == null) {
			return null;
		}

		JunOpenGL3dObject newDisplayObject = terrainField.asJunOpenGL3dObject();
		newDisplayObject.paint_(Color.green);

		return newDisplayObject;
	}

	/**
	 * Answer the default controller.
	 * 
	 * @return sra.smalltalk.StController
	 */
	protected StController defaultController() {
		return new JunTerrainDisplayController();
	}

	/**
	 * Answer the default far.
	 * 
	 * @return double
	 */
	protected double defaultFar() {
		return this.defaultEyePoint().distance_(this.defaultSightPoint()) * this.defaultViewFactor();
	}

	/**
	 * Create the menubar for JunOpenGLDisplay.
	 * 
	 * @return sra.smalltalk.StMenuModel
	 */
	protected StMenuModel defaultMenu() {
		StMenuModel rootMenu = new StMenuModel();

		rootMenu.add(this._fileMenu());
		rootMenu.add(this._viewMenu());
		rootMenu.add(this._lightMenu());
		rootMenu.add(this._miscMenu());

		return rootMenu;
	}

	/**
	 * Answer the default near.
	 * 
	 * @return double
	 */
	protected double defaultNear() {
		return this.defaultFar() / 10000.0d;
	}

	/**
	 * Answer the height of the eye at the specified point.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @return double
	 */
	protected double eyeHeightAtPoint_(Jun2dPoint aPoint) {
		double z = terrainField.zAt2dPoint_(aPoint);

		if (Double.isNaN(z)) {
			z = 0.0d;
		}

		return z + this.inches();
	}

	/**
	 * Grab and rotate the 3d object.
	 * 
	 * @param from2dPoint sra.jun.grometry.basic.Jun2dPoint
	 * @param to2dPoint sra.jun.grometry.basic.Jun2dPoint
	 */
	protected void grab_x_(Jun2dPoint from2dPoint, Jun2dPoint to2dPoint) {
		if (this.displayObject() == null) {
			return;
		}

		JunOpenGLProjector projector = this.displayProjector();
		JunAngle rotationAngle = JunAngle.FromRad_((to2dPoint.x() - from2dPoint.x()) / 2.0d);

		if (Math.abs(rotationAngle.rad()) > ThetaAccuracy) {
			Jun3dLine rotationAxis = new Jun3dLine(projector.eyePoint(), projector.eyePoint().plus_(new Jun3dPoint(0, 0, 1)));
			Jun3dTransformation transformationInv = Jun3dTransformation.Rotate_around_(rotationAngle.mul_(-1), rotationAxis);
			projector.sightPoint_(transformationInv.applyTo_(projector.sightPoint()));
			projector.upVector_(new Jun3dPoint(0, 0, 1));
		}
	}

	/**
	 * Grab and rotate the 3d object.
	 * 
	 * @param from2dPoint sra.jun.grometry.basic.Jun2dPoint
	 * @param to2dPoint sra.jun.grometry.basic.Jun2dPoint
	 */
	protected void grab_y_(Jun2dPoint from2dPoint, Jun2dPoint to2dPoint) {
		if (this.displayObject() == null) {
			return;
		}

		JunOpenGLProjector projector = this.displayProjector();
		JunOpenGLProjection projection = projector.projection();
		JunAngle rotationAngle = JunAngle.FromRad_((to2dPoint.y() - from2dPoint.y()) / 2.0d);

		if (Math.abs(rotationAngle.rad()) > ThetaAccuracy) {
			Jun3dLine rotationAxis = new Jun3dLine(projection.eyePoint(), projection.eyePoint().plus_(projection.rightVector()));
			Jun3dTransformation transformationInv = Jun3dTransformation.Rotate_around_(rotationAngle, rotationAxis);
			projector.sightPoint_(transformationInv.applyTo_(projection.sightPoint()));
			projector.upVector_(new Jun3dPoint(0, 0, 1));
		}
	}

	/**
	 * Initialize the JunTerrainDisplayModel.
	 */
	protected void initialize() {
		super.initialize();

		terrainField = null;
		inches = 1.0d;
		defaultZoomHeight = 1.0d;
	}

	/**
	 * Move your eyes to the specified point.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun2dPoint
	 */
	protected void moveTo_(Jun2dPoint aPoint) {
		JunOpenGLProjector projector = this.displayProjector();
		Jun3dPoint currentEyePoint = projector.eyePoint();
		Jun3dPoint newEyePoint = new Jun3dPoint(aPoint, this.eyeHeightAtPoint_(aPoint));
		Jun3dPoint newSightPoint = projector.sightPoint().plus_(newEyePoint).minus_(currentEyePoint);
		projector.sightPoint_(newSightPoint);
		projector.eyePoint_(newEyePoint);
		projector.upVector_(new Jun3dPoint(0, 0, 1));
		this.changed_($("projection"));
	}

	/**
	 * Move your eyes to the specified point.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun3dPoint
	 */
	protected void moveTo_(Jun3dPoint aPoint) {
		this.moveTo_(new Jun2dPoint(aPoint.x(), aPoint.y()));
	}

	/**
	 * Answer the window title.
	 * 
	 * @return java.lang.String
	 */
	protected String windowTitle() {
		return JunSystem.$String("Terrain");
	}
}
