package sra.smalltalk;

import java.util.Locale;
import java.util.MissingResourceException;
import java.util.ResourceBundle;

/**
 * StMessageCatalog class
 * 
 *  @author    nisinaka
 *  @created   2004/03/18 (by nisinaka)
 *  @updated   N/A
 *  @version   6.5
 *  @copyright 1999-2004 SRA (Software Research Associates, Inc.)
 *  @copyright 2001-2004 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: StMessageCatalog.java,v 6.1 2004/04/15 06:28:22 nisinaka Exp $
 */
public class StMessageCatalog extends StObject {

	protected String bundleName;
	protected Locale locale;
	protected ResourceBundle resourceBundle;

	static {
		// Bug Id : 4152725
		Locale defaultLocale = Locale.getDefault();
		String language = System.getProperty("user.language");
		if (defaultLocale.getLanguage().equals(language) == false) {
			String country = System.getProperty("user.country", "");
			String variant = System.getProperty("user.variant", "");
			if (country.length() == 0) {
				String region = System.getProperty("user.region");
				if (region != null) {
					int index = region.indexOf("_");
					if (index >= 0) {
						country = region.substring(0, index);
						variant = region.substring(index + 1);
					} else {
						country = region;
					}
				}
			}
			Locale.setDefault(new Locale(language, country, variant));
		}
	}

	/**
	 * Create a new instance of StMessageCatalog and initialize it.
	 * 
	 * @param aBundleName java.lang.String
	 * @param aLocale java.util.Locale
	 * @category Instance creation
	 */
	public StMessageCatalog(String aBundleName) {
		this(aBundleName, Locale.getDefault());
	}

	/**
	 * Create a new instance of StMessageCatalog and initialize it.
	 * 
	 * @param aBundleName java.lang.String
	 * @param aLocale java.util.Locale
	 * @category Instance creation
	 */
	public StMessageCatalog(String aBundleName, Locale aLocale) {
		bundleName = aBundleName;
		this.setLocale(aLocale);
	}

	/**
	 * Set the new locale.
	 * 
	 * @param aLocale java.util.Locale
	 * @category accessing
	 */
	public void setLocale(Locale aLocale) {
		locale = aLocale;

		if (locale == null) {
			resourceBundle = null;
		} else {
			try {
				resourceBundle = ResourceBundle.getBundle(bundleName, locale);
			} catch (MissingResourceException e) {
				resourceBundle = null;
			}
		}
	}

	/**
	 * Look up the string for the key from the message catalog.
	 * 
	 * @param key java.lang.String
	 * @return java.lang.String
	 * @category accessing
	 */
	public String lookUp(String key) {
		return this.lookUp(key, null);
	}

	/**
	 * Look up the string for the key from the message catalog.
	 * 
	 * @param key java.lang.String
	 * @param defaultString java.lang.String
	 * @return java.lang.String
	 * @category accessing
	 */
	public String lookUp(String key, String defaultString) {
		if (defaultString == null) {
			defaultString = key;
		}

		if (resourceBundle == null) {
			return this.lookUpFailed(key, defaultString);
		} else {
			try {
				return resourceBundle.getString(key);
			} catch (MissingResourceException e) {
				return this.lookUpFailed(key, defaultString);
			}
		}
	}

	/**
	 * Look up the string for the key defined in the resource bundle.
	 * Parameter substitutions on the string will be done.
	 * 
	 * @param key java.lang.String
	 * @param defaultString java.lang.String
	 * @param parameters java.lang.Object
	 * @return java.lang.String
	 * @category accessing
	 */
	public String lookUp(String key, String defaultString, Object parameter) {
		return this.lookUp(key, defaultString, new Object[] { parameter });
	}

	/**
	 * Look up the string for the key defined in the resource bundle.
	 * Parameter substitutions on the string will be done.
	 * I wish I could use String#replaceAll()...
	 * 
	 * @param key java.lang.String
	 * @param defaultString java.lang.String
	 * @param parameters java.lang.Object[]
	 * @return java.lang.String
	 * @category accessing
	 */
	public String lookUp(String key, String defaultString, Object[] parameters) {
		String string = this.lookUp(key, defaultString);
		if (parameters != null) {
			for (int i = 0; i < parameters.length; i++) {
				StringBuffer stringBuffer = new StringBuffer(string.length());
				String pkey = "<" + (i + 1) + "p>";
				int from = 0, index;
				while ((index = string.indexOf(pkey, from)) >= 0) {
					stringBuffer.append(string.substring(from, index));
					stringBuffer.append(parameters[i].toString());
					from = index + pkey.length();
				}
				stringBuffer.append(string.substring(from));
				string = stringBuffer.toString();
			}
		}
		return string;
	}

	/**
	 * Called when the message for the key is missing.
	 * The subclasses may override this to handle exception in other way. 
	 * 
	 * @param key java.lang.String
	 * @param defaultString java.lang.String
	 * @return java.lang.String
	 * @category private
	 */
	protected String lookUpFailed(String key, String defaultString) {
		return defaultString;
	}

}
